/**
 * Copyright (c) 2012 Research In Motion Limited.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at:
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global window, document, console, cc, ccCanvas, ControlsOverlay */

var CocosApp = cc.LayerColor.extend({
	controls: null, /* Will be used to instantiate a ControlsOverlay Node. */

    init: function () {
		'use strict';
		var layer, movement, button, colour;

        /* Our Layer. */
        this._super();
        this.initWithColor(new cc.Color4B(50, 50, 50, 255));
		layer = this;

		/* Create a new ControlsOverlay object. */
		this.controls = new ControlsOverlay({
			canvas: ccCanvas
		});

		/* Add a Joystick. We will use this to move the button around the screen. */
		movement = this.controls.addJoystick({
			imageBase: './Resources/controls_overlay/dpad.png',
			imagePad: './Resources/controls_overlay/pad.png',
			fixed: false,
			trigger: new cc.Rect(0, 0, ccCanvas.width / 2.0, ccCanvas.height),
			opacLow: 0.0
		});

		/* Add a Joystick. We will use this to change the colour of the button. */
		colour = this.controls.addJoystick({
			imageBase: './Resources/controls_overlay/dpad.png',
			imagePad: './Resources/controls_overlay/pad.png',
			fixed: true,
			pos: new cc.Point(ccCanvas.width - 100.0, 100.0),
			opacLow: 80.0
		});

		/* Add a Button. We add it last so that it is rendered on top of other controls. */
		button = this.controls.addButton({
			image: './Resources/controls_overlay/buttonred.png',
			pos: new cc.Point(ccCanvas.width - 54.0, ccCanvas.height - 54.0),
			opacLow: 80.0
		});

		/* Update the Button's position based on the movement Joystick. We'll schedule an update every frame. */
		button.update = function (dt) {
			var position;

			/* Calculate the time since last frame to normalize movement speed (at 60 fps we expect 0.017 seconds per frame), and multiply by an arbitrary scaling factor. */
			dt = dt / 0.017 * 0.15;

			/* Get the current position of the Button. */
			position = this.getPosition();

			/* Update our position based on the velocity of the movement Joystick. We'll keep a 54.0 pixel border that the button can not cross. */
			position.x = Math.max(54.0, Math.min(position.x + movement.velocity.x * dt, ccCanvas.width - 54.0));
			position.y = Math.max(54.0, Math.min(position.y + movement.velocity.y * dt, ccCanvas.height - 54.0));

			/* Update our Button's position and trigger area. */
			button.setPosition(position);
			button.trigger = button.getBoundingBox();
		};

		/* Start updating our Button's position every frame. */
		button.scheduleUpdate();

		/* When the Button is tapped, we'll change the background colour. */
		button.color = new cc.Color4B(120, 0, 0, 255);
		button.onTouchEnd = function (touch, point) {
			if (cc.Rect.CCRectContainsPoint(this.trigger, point)) {
				layer.setColor(this.color);
			}
		};

		/* Our colour Joystick will update the image used for the Button. To minimize work, we'll keep track of the previous Joystick direction. */
		colour.prevDirection = -1;
		colour.onTouchMove = function () {
			/* If this Joystick's direction has changed since our previous direction... */
			if (this.prevDirection !== this.direction) {
				/* ...update to the new direction. */
				this.prevDirection = this.direction;

				/* We are ignoring diagonals, and only updating with a 4-axis system. We could easily add the omitted directions though. */
				if (this.direction === 0) {
					/* When this Joystick is positioned West, we will have a red Button. */
					button.setTexture(cc.TextureCache.getInstance().textureForKey('./Resources/controls_overlay/buttonred.png'));
					button.color = new cc.Color4B(120, 0, 0, 255);
				} else if (this.direction === 2) {
					/* When this Joystick is positioned North, we will have a green Button. */
					button.setTexture(cc.TextureCache.getInstance().textureForKey('./Resources/controls_overlay/buttongreen.png'));
					button.color = new cc.Color4B(0, 120, 0, 255);
				} else if (this.direction === 4) {
					/* When this Joystick is positioned East, we will have a blue Button. */
					button.setTexture(cc.TextureCache.getInstance().textureForKey('./Resources/controls_overlay/buttonblue.png'));
					button.color = new cc.Color4B(0, 0, 120, 255);
				} else if (this.direction === 6) {
					/* When this Joystick is positioned South, we will have a yellow Button. */
					button.setTexture(cc.TextureCache.getInstance().textureForKey('./Resources/controls_overlay/buttonyellow.png'));
					button.color = new cc.Color4B(120, 120, 0, 255);
				}
			}
		};

		/* Add our ControlsOverlay Node to our LayerColor. */
		this.addChild(this.controls);

        return true;
    }
});

var CocosAppScene = cc.Scene.extend({
    onEnter: function () {
		'use strict';

        this._super();

        var layer = new CocosApp();
        layer.init();

        this.addChild(layer);
    }
});
